/***************************************************************************
                          MatrixAdditon.cpp  -  description
                             -------------------
    begin                : 2015/02/04
    copyright            : (C) 2015 by Tomáš Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>
#include <cmath>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <TimerRT.h>

using namespace std;

template< typename Real >
class StdMatrix
{
   public:
      StdMatrix( int size )
      {
         this->matrix.resize( size );
         for( int i = 0; i < size; i++ )
            this->matrix[ i ].resize( size );
         this->size = size;
      };

      void setValue( const Real& v )
      {
         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
               this->matrix[ i ][ j ] = v;
      }

      void addMatrixRowWise( const StdMatrix< Real >& m )
      {
         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
               this->matrix[ i ][ j ] += m.matrix[ i ][ j ];
      }

      void addMatrixColumnWise( const StdMatrix< Real >& m )
      {
         for( int j = 0; j < size; j++ )
            for( int i = 0; i < size; i++ )
               this->matrix[ i ][ j ] += m.matrix[ i ][ j ];
      }


   protected:
      vector< vector< Real > > matrix;

      int size;
};

template< typename Real >
class ArrayMatrix
{
   public:
      ArrayMatrix( int size )
      {
         this->matrix = new Real[ size * size ];
         this->size = size;
      }

      void setValue( const Real& v )
      {
         for( int i = 0; i < size * size; i++ )
            this->matrix[ i ] = v;
      }

      void addMatrixRowWise( const ArrayMatrix< Real >& m )
      {
         for( int i = 0; i < size; i++ )
            for( int j = 0; j < size; j++ )
               this->matrix[ i * size + j ] += m.matrix[ i * size + j ];
      }

      void addMatrixColumnWise( const ArrayMatrix< Real >& m )
      {
         for( int j = 0; j < size; j++ )
            for( int i = 0; i < size; i++ )
               this->matrix[ i * size + j ] += m.matrix[ i * size + j ];
      }


      ~ArrayMatrix()
      {
         delete[] this->matrix;
      }
   protected:
      Real* matrix;

      int size;
};

void writeTableHeader( ostream& str )
{
   str << std::right;
   str << "#";
   str << setw( 140 ) << std::setfill( '-') << "-" << endl;
   str << "#";
   str << std::setfill( ' ' );
   str << setw( 40 ) << "Matrix size         "
       << setw( 40 ) << "STD Matrix          "
       << setw( 40 ) << "Array Matrix        " << endl;
   str << "#";
   str << setw( 40 ) << ""
       << setw( 20 ) << "Row wise"
       << setw( 20 ) << "Column wise"
       << setw( 20 ) << "Row wise"
       << setw( 20 ) << "Column wise" << endl;
   str << "#";
   str << setw( 20 ) << "size"
       << setw( 20 ) << "log2"
       << setw( 20 ) << "GFLOPS"
       << setw( 20 ) << "GFLOPS"
       << setw( 20 ) << "GFLOPS"
       << setw( 20 ) << "GFLOPS" << endl;
   str << "#";
   str << setw( 140 ) << std::setfill( '-') << "-" << endl;
   str << std::setfill( ' ' );
}

void writeTableLine( ostream& str,
                     const int size,
                     const double& stdMatrixRowWiseGlops,
                     const double& stdMatrixColumnWiseGlops,
                     const double& arrayMatrixRowWiseGlops,
                     const double& arrayMatrixColumnWiseGlops )
{
   str << setw( 20 ) << size
       << setw( 20 ) << log2( size )
       << setw( 20 ) << stdMatrixRowWiseGlops
       << setw( 20 ) << stdMatrixColumnWiseGlops
       << setw( 20 ) << arrayMatrixRowWiseGlops
       << setw( 20 ) << arrayMatrixColumnWiseGlops
       << endl;
}

int main( int argc, char* argv[] )
{
   const long int maxSize = 16384;
   const double minTestTime( 1.0 );
   TimerRT timer;

   fstream outputFile;
   outputFile.open( "results.txt", ios::out );
   if( ! outputFile )
   {
      cerr << "I cannot open the output file results.txt." << endl;
      return EXIT_FAILURE;
   }

   writeTableHeader( cout );
   writeTableHeader( outputFile );
   int size = 16;
   while( size < maxSize )
   {
      const long int elements = size * size;
      const int repeating = maxSize * maxSize / elements;
      StdMatrix< float > stdMatrix1( size ), stdMatrix2( size );
      ArrayMatrix< float > arrayMatrix1( size ), arrayMatrix2( size );
      stdMatrix1.setValue( 0.0 );
      stdMatrix2.setValue( 1.0 );
      arrayMatrix1.setValue( 0.0 );
      arrayMatrix2.setValue( 1.0 );
      int counter;

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         stdMatrix1.addMatrixRowWise( stdMatrix2 );
         counter++;
      }
      timer.stop();
      double time = timer.getTime();
      double stdMatrixRowWiseGflops = elements * counter / time / 1.0e9;

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         stdMatrix1.addMatrixColumnWise( stdMatrix2 );
         counter++;
      }
      timer.stop();
      time = timer.getTime();
      double stdMatrixColumnWiseGflops = elements * counter / time / 1.0e9;


      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         arrayMatrix1.addMatrixColumnWise( arrayMatrix2 );
         counter++;
      }
      timer.stop();
      time = timer.getTime();
      double arrayMatrixColumnWiseGflops = elements * counter / time / 1.0e9;

      counter = 0;
      timer.reset();
      timer.start();
      while( timer.getTime() < minTestTime )
      {
         arrayMatrix1.addMatrixRowWise( arrayMatrix2 );
         counter++;
      }
      timer.stop();
      time = timer.getTime();
      double arrayMatrixRowWiseGflops = elements * counter / time / 1.0e9;

      writeTableLine( cout,
                      size,
                      stdMatrixRowWiseGflops,
                      stdMatrixColumnWiseGflops,
                      arrayMatrixRowWiseGflops,
                      arrayMatrixColumnWiseGflops );
      writeTableLine( outputFile,
                      size,
                      stdMatrixRowWiseGflops,
                      stdMatrixColumnWiseGflops,
                      arrayMatrixRowWiseGflops,
                      arrayMatrixColumnWiseGflops );


      size *= 2;
   }
}

