/***************************************************************************
                          MatrixTransposition.cpp  -  description
                             -------------------
    begin                : 2015/02/04
    copyright            : (C) 2015 by Tomáš Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>
#include <cstring>
#include <cmath>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <TimerRT.h>

using namespace std;

template< typename Real >
class ArrayMatrix
{
   public:
      ArrayMatrix( int size )
      {
         this->matrix = new Real[ size * size ];
         //posix_memalign( ( void** ) &this->matrix, 64, size * size * sizeof( Real ) );
         this->size = size;
      }

      void setValue( const Real& v )
      {
         for( int i = 0; i < size * size; i++ )
            this->matrix[ i ] = v;
      }

      void setElement( int i, int j, const Real& v )
      {
         this->matrix[ i * size + j ] = v;
      }

      void transpose()
      {
         Real tmp;
         for( int i = 1; i < size; i++ )
            for( int j = 0; j < i; j++ )
            {
               tmp = this->matrix[ i * size + j ];
               this->matrix[ i * size + j ] = this->matrix[ j * size + i ];
               this->matrix[ j * size + i ] = tmp;
            }
      }

      ~ArrayMatrix()
      {
         delete[] this->matrix;
      }
   protected:
      Real* matrix;

      int size;
};

void writeTableHeader( ostream& str )
{
   str << std::right;
   str << "#";
   str << setw( 80 ) << std::setfill( '-') << "-" << endl;
   str << "#";
   str << std::setfill( ' ' );
   str << setw( 40 ) << "Matrix size       "
       << setw( 40 ) << "Transposition"
       << endl;
   str << "#";
   str << setw( 20 ) << "size"
       << setw( 20 ) << "log2"
       << setw( 20 ) << "s"
       << setw( 20 ) << "GFLOPS"
       << endl;
   str << "#";
   str << setw( 80 ) << std::setfill( '-') << "-" << endl;
   str << std::setfill( ' ' );
}

void writeTableLine( ostream& str,
                     const int size,
                     const double& time,
                     const double& gflops )
{
   str << setw( 20 ) << size
       << setw( 20 ) << log2( size )
       << setw( 20 ) << time
       << setw( 20 ) << gflops
       << endl;
}
void writeTableDelimiter( ostream& str )
{
   str << "#";
   str << setw( 80 ) << std::setfill( '-') << "-" << endl;
   str << std::setfill( ' ' );
}

void testTransposition( const int size,
                        fstream& outputFile )
{
   const double minTestTime( 1.0 );
   TimerRT timer;

   ArrayMatrix< double > a( size );
   a.setValue( 1.0 );

   int counter;

   const double operations = 0.5 * size * size;
   counter = 0;
   timer.reset();
   timer.start();
   while( timer.getTime() < minTestTime )
   {
      a.transpose();
      counter++;
   }
   timer.stop();
   double time = timer.getTime() / ( double ) counter;
   double gflops = operations / time / 1.0e9;

   writeTableLine( cout,
                   size,
                   time,
                   gflops );
   writeTableLine( outputFile,
                   size,
                   time,
                   gflops );
}

int main( int argc, char* argv[] )
{
   fstream outputFile;
   outputFile.open( "results.txt", ios::out );
   if( ! outputFile )
   {
      cerr << "I cannot open the output file results.txt." << endl;
      return EXIT_FAILURE;
   }

   writeTableHeader( cout );
   writeTableHeader( outputFile );


   testTransposition( 15, outputFile );
   testTransposition( 16, outputFile );
   testTransposition( 17, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 31, outputFile );
   testTransposition( 32, outputFile );
   testTransposition( 33, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 63, outputFile );
   testTransposition( 64, outputFile );
   testTransposition( 65, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 127, outputFile );
   testTransposition( 128, outputFile );
   testTransposition( 129, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 255, outputFile );
   testTransposition( 256, outputFile );
   testTransposition( 257, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 511, outputFile );
   testTransposition( 512, outputFile );
   testTransposition( 513, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 1023, outputFile );
   testTransposition( 1024, outputFile );
   testTransposition( 1025, outputFile );

   writeTableDelimiter( cout );
   writeTableDelimiter( outputFile );

   testTransposition( 2047, outputFile );
   testTransposition( 2048, outputFile );
   testTransposition( 2049, outputFile );


}

